/*
 * $RCSfile: ppp_fsm.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1994  Arizona Board of Regents
 *
 * $Log: ppp_fsm.c,v $
 * Revision 1.2  1994/10/27 17:27:28  andrey
 * removed non-ansi definitions
 *
 * Revision 1.1  1994/10/24  20:44:34  andrey
 * Initial revision
 *
 *
 * $Revision: 1.2 $
 * $Date: 1994/10/27 17:27:28 $
 */

static void process_timeout(Event self, void * fip);
void ppp_error(FSM_info *);

static char * event_name[last_event+1] = {
  "UP",
  "DOWN",
  "OPEN",
  "CLOSE",
  "TO+",
  "TO-",
  "RCR+",
  "RCR-",
  "RCA",
  "RCN",
  "RTR",
  "RTA",
  "RUC",
  "RXJ+",
  "RXJ-",
  "RXR",
};

/*
 * state attribute table
 */

static u_char state_attrs[last_state] = {
	TIMER_OFF,	/* F_initial */
	TIMER_OFF,	/* F_starting, */
	TIMER_OFF,	/* F_closed, */
	TIMER_OFF,	/* F_stopped, */

	TIMER_ON,	/* F_closing, */
	TIMER_ON,	/* F_stopping, */
	TIMER_ON,	/* F_req_sent, */
	TIMER_ON,	/* F_ack_rcvd, */
	TIMER_ON,	/* F_ack_sent, */

	TIMER_OFF,	/* F_opened, */
};

static char * state_name[last_state] = {
	"Initial",	/* F_initial */
	"Starting",	/* F_starting, */
	"Closed",	/* F_closed, */
	"Stopped",	/* F_stopped, */
	"Closing",	/* F_closing, */
	"Stopping",	/* F_stopping, */
	"Req_Sent",	/* F_req_sent, */
	"Ack_Rcvd",	/* F_ack_rcvd, */
	"Ack_Sent",	/* F_ack_sent, */
	"Opened",	/* F_opened, */
};

/* ------------------------------------------------------------------------ */
/*                       automaton support routines                         */
/* ------------------------------------------------------------------------ */

/*
 * turn off timeout, if it's still outstanding
 * release the event
 */
static void
cancel_timer(fi)
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP0(sess, TR_DETAILED, "cancel timer");

  if (fi->timerev) {                /* release event */
    /* do something useful with ecr */
    /*
    EvCancelReturn ecr = evCancel(fi->timerev);
     */
    evDetach(fi->timerev);
    fi->timerev = 0;
  }
}

static void
startup_timer(fi)
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP0(sess, TR_DETAILED, "startup timer");

  fi->timerev = evSchedule(process_timeout, (void *)fi,
                               fi->time_options->timeval);
  fi->restart_counter--;
}

/* ------------------------------------------------------------------------ */

/*
 * grind through the automaton
 */
void
ppp_process_event(fi, pe)
  FSM_info * fi;
  fsm_event pe;
{
  XObj sess = fi->session;

  fsm_state prev_state = fi->state;

  fsm_act action = (*(fi->action_table))[pe][prev_state];
  fsm_state next_state = (*(fi->state_table))[pe][prev_state];

  xTraceP0(sess, TR_DETAILED, "process event");

  if (pe == NO_EVENT) {
    xTraceP1(sess, TR_DETAILED, "state trans: in %s, received NOEVENT",
       state_name[prev_state]);

    if (fi->msg) {
      msgDestroy(fi->msg);
      fi->msg = 0;
    }
    return;
  }

  if (next_state == F_err) {
    action = ppp_error;
    next_state = prev_state;
  }

  xTraceP4(sess, TR_DETAILED,
     "state transition: in %s, received event %s %s going to %s",
       state_name[prev_state],
       event_name[pe],
       (next_state == F_err) ? "(error)" : "",
       state_name[next_state]);

  if (action) (*action)(fi);

  if (fi->msg) {
    msgDestroy(fi->msg);
    fi->msg = 0;
  }

  /* per RFC 1548: turn off timer when "transitioning"
   * from state accepting timer events to one not accepting
   */
  if ((state_attrs[prev_state] & TIMER_ON) &&
      (state_attrs[next_state] & TIMER_OFF)) {
     cancel_timer(fi);
  }

  fi->state = next_state;
}

/* ------------------------------------------------------------------------ */
/*
 * field the event timeout, and kick the automaton again
 */
static void
process_timeout(self, fip)
  Event self;
  void * fip;
{
  FSM_info * fi = (FSM_info *)fip;
  XObj sess = fi->session;

  fsm_event next_event = NO_EVENT;

  xTraceP0(sess, TR_DETAILED, "process timeout");
  if (fi->restart_counter > 0) {
    next_event = F_timeout_pos;
  }
  else {
    next_event = F_timeout_zero;
  }
  ppp_process_event(fi, next_event);
}
/* ------------------------------------------------------------------------ */
/*                           automaton actions                              */
/* ------------------------------------------------------------------------ */

/* --------------- actions that communicate with outer layers ------------- */
/*
 * this layer up
 */
void
ppp_tlu(fi)
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP0(sess, TR_DETAILED, "tlu action");
  fi->nego_ok = 1;
  if (fi->waiting_on_nego) {
    /*
     * squirrel state info away to inform waiter
     */
    semSignal(&fi->Nego_sem);
  }
  else {
    fi->signal_negotiation_done(sess);
  }
}

/*
 * this layer finished
 *   negotiation failed
 */
void
ppp_tlf(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP0(sess, TR_DETAILED, "tlf action");
  fi->nego_ok = 0;
  if (fi->waiting_on_nego) {
    /*
     * squirrel state info away to inform waiter
     */
    semSignal(&fi->Nego_sem);
  }
  else {
    fi->signal_negotiation_done(sess);
  }
}

/*
 * this layer down
 */
void
ppp_tld(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP0(sess, TR_DETAILED, "tld action");
}

/*
 * this layer started
 *   ignore, since this is generated by OPEN and UP anyway
 */
void
ppp_tls(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP0(sess, TR_DETAILED, "tls action");
}

/* ------------------------------------------------------------------------ */

/*
 * initialize restart counter
 */
void
ppp_irc(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  int n = 0;

  xTraceP0(sess, TR_DETAILED, "irc action");
  switch(fi->state) {
    case F_starting:
    case F_closed:
    case F_stopped:
      n = fi->time_options->config_retry;
    break;

    case F_req_sent:
    case F_ack_rcvd:
    case F_ack_sent:
    case F_opened:
      n = fi->time_options->term_retry;
    break;

    case F_initial:
    case F_closing:
    case F_stopping:
    default:
      xTraceP0(sess, TR_ERRORS, "bad state in irc action");
  }
  fi->restart_counter = n;
  cancel_timer(fi);
}

/*
 * zero restart counter
 *  also restarts the clock
 */
void
ppp_zrc(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP0(sess, TR_DETAILED, "zrc action");
  fi->restart_counter = 0;
  startup_timer(fi);
}

/*
 * send configure request
 */
void
ppp_scr(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;
  PPP_header lh;
  Msg * msg;

  xTraceP0(sess, TR_DETAILED, "scr action");
  PPP_CONF_REQ_init(&lh);
  lh.ident = 0;
  msg = fi->make_config_req_msg(sess);
  lh.length = msgLen(msg) + sizeof(PPP_header);
  msgPush(msg, PPP_hdr_store, &lh, sizeof(PPP_header), 0);
  xPush(xGetDown(sess, 0), msg);
  msgDestroy(msg);
  startup_timer(fi);
}

/*
 * send configure ack
 */
void
ppp_sca(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  PPP_header lh;
  Msg * msg;

  xTraceP0(sess, TR_DETAILED, "sca action");
  PPP_CONF_ACK_init(&lh);
  lh.ident = 0;
  msg = fi->make_config_ack_msg(sess);
  lh.length = msgLen(msg) + sizeof(PPP_header);
  msgPush(msg, PPP_hdr_store, &lh, sizeof(PPP_header), 0);
  xPush(xGetDown(sess, 0), msg);
  msgDestroy(msg);
}

/*
 * send configure nak
 */
void
ppp_scn(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;
  PPP_header lh;
  Msg * msg;

  xTraceP0(sess, TR_DETAILED, "scn action");
  lh.ident = 0;
  if (fi->nak_not_rej) {
    PPP_CONF_NAK_init(&lh);
    msg = fi->make_config_nak_msg(sess);
  }
  else {
    PPP_CONF_REJ_init(&lh);
    msg = fi->make_config_rej_msg(sess);
  }
  lh.length = msgLen(msg) + sizeof(PPP_header);
  msgPush(msg, PPP_hdr_store, &lh, sizeof(PPP_header), 0);
  xPush(xGetDown(sess, 0), msg);
  msgDestroy(msg);
}

/*
 * send terminate request
 */
void
ppp_str(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;
  PPP_header lh;
  Msg msg;

  xTraceP0(sess, TR_DETAILED, "str action");
  PPP_TERM_REQ_init(&lh);
  lh.ident = 0;
  msgConstructEmpty(&msg);
  lh.length = msgLen(&msg) + sizeof(PPP_header);
  msgPush(&msg, PPP_hdr_store, &lh, sizeof(PPP_header), 0);
  xPush(xGetDown(sess, 0), &msg);
  msgDestroy(&msg);
  startup_timer(fi);
}

/*
 * send terminate ack
 */
void
ppp_sta(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  PPP_header lh;
  Msg msg;

  xTraceP0(sess, TR_DETAILED, "sta action");
  PPP_TERM_ACK_init(&lh);
  lh.ident = 0;
  msgConstructEmpty(&msg);
  lh.length = msgLen(&msg) + sizeof(PPP_header);
  msgPush(&msg, PPP_hdr_store, &lh, sizeof(PPP_header), 0);
  xPush(xGetDown(sess, 0), &msg);
  msgDestroy(&msg);
}

/*
 * send code reject
 */
void
ppp_scj(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP0(sess, TR_DETAILED, "scj action");
}

/*
 * send echo reply
 */
void
ppp_ser(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP0(sess, TR_DETAILED, "ser action");
}

/* irc, then scr */
void
ppp_irc_scr(fi) 
  FSM_info * fi;
{
  ppp_irc(fi); ppp_scr(fi);
}

/* irc, then scr, then sca */
void
ppp_irc_scr_sca(fi) 
  FSM_info * fi;
{
  ppp_irc(fi); ppp_scr(fi); ppp_sca(fi);
}

/* irc, then scr, then scn */
void
ppp_irc_scr_scn(fi) 
  FSM_info * fi;
{
  ppp_irc(fi); ppp_scr(fi); ppp_scn(fi);
}

void
ppp_irc_str(fi) 
  FSM_info * fi;
{
  ppp_irc(fi); ppp_str(fi);
}

void
ppp_tld_irc_str(fi) 
  FSM_info * fi;
{
  ppp_tld(fi); ppp_irc(fi); ppp_str(fi);
}

void
ppp_tld_scr_sca(fi) 
  FSM_info * fi;
{
  ppp_tld(fi); ppp_scr(fi); ppp_sca(fi);
}

void
ppp_tld_scr_scn(fi) 
  FSM_info * fi;
{
  ppp_tld(fi); ppp_scr(fi); ppp_scn(fi);
}

void
ppp_irc_tlu(fi) 
  FSM_info * fi;
{
  ppp_irc(fi); ppp_tlu(fi);
}

void
ppp_tld_scr(fi) 
  FSM_info * fi;
{
  ppp_tld(fi); ppp_scr(fi);
}

void
ppp_tld_zrc_sta(fi) 
  FSM_info * fi;
{
  ppp_tld(fi); ppp_zrc(fi); ppp_sta(fi);
}

void
ppp_sca_tlu(fi) 
  FSM_info * fi;
{
  ppp_sca(fi); ppp_tlu(fi);
}

void
ppp_error(fi) 
  FSM_info * fi;
{
  XObj sess = fi->session;

  xTraceP1(sess, TR_DETAILED, "bad state transition - in state %s",
    state_name[fi->state]);
}

/* ------------------------------------------------------------------------ */
/*                            header routines                               */
/* ------------------------------------------------------------------------ */
void
PPP_hdr_store(src, dest, count, type)
  void * src;
  char * dest;
  long count;
  void * type;
{
  u_short n;
  ((PPP_header*)dest)->code = ((PPP_header*)src)->code;
  ((PPP_header*)dest)->ident = ((PPP_header*)src)->ident;
  n = htons(((PPP_header*)src)->length);
  bcopy((char *)&n, (char *)&((PPP_header*)dest)->length, sizeof(n));
}

long
PPP_hdr_load(dest, src, count, type)
  void * dest;
  char * src;
  long count;
  void * type;
{
  u_short n;
  ((PPP_header*)dest)->code = ((PPP_header*)src)->code;
  ((PPP_header*)dest)->ident = ((PPP_header*)src)->ident;

  bcopy((char *)&((PPP_header*)src)->length, (char *)&n, sizeof(n));
  ((PPP_header*)dest)->length = ntohs(n);
  return sizeof(PPP_header);
}

int
ppp_decode_packet(msg, id, len)
  Msg * msg;
  int * id;
  int * len;
{
  PPP_header lh;

  msgPop(msg, PPP_hdr_load, &lh, sizeof(PPP_header), 0);
  *id = lh.ident;
  *len = lh.length;
  return lh.code;
}


void
PPP_opt_hdr_store(src, dest, count, type)
  void * src;
  char * dest;
  long count;
  void * type;
{
  ((PPP_opt*)dest)->type = ((PPP_opt*)src)->type;
  ((PPP_opt*)dest)->length = ((PPP_opt*)src)->length;
}

long
PPP_opt_hdr_load(dest, src, count, type)
  void * dest;
  char * src;
  long count;
  void * type;
{
  ((PPP_opt*)dest)->type = ((PPP_opt*)src)->type;
  ((PPP_opt*)dest)->length = ((PPP_opt*)src)->length;
  return sizeof(PPP_opt);
}


