/*
 * $RCSfile: tcp_x.c,v $
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Log: tcp_x.c,v $
 * Revision 1.1  1996/07/11 21:39:21  nicko
 * Initial revision
 *
 * Revision 1.55.1.8  1995/11/04  00:01:54  ho
 * Added code for dealing with multiple saids on the security participant.
 *
 * Revision 1.55.1.7  1995/11/03  18:00:02  ho
 * Fixed security participants copy operation.  Had been reversing order.
 *
 * Revision 1.55.1.6  1995/06/07  21:43:31  spatsch
 * added ipsec support.
 *
 * Revision 1.55.1.5  1994/09/20  18:52:30  gkim
 * added port range support (ANY_LOW_PORT, ANY_HIGH_PORT)
 *
 * Revision 1.55.1.4  1994/08/08  03:24:04  ho
 * Made local host addr a per session calculation, not per protocol.
 *
 * Revision 1.55.1.3  1994/06/30  21:43:44  davidm
 * (tcpOpen): removed extraneous tcp_destroy() call---when waking up
 * with t_state != TCPS_ESTABLISHED, tcp_drop() has destroyed the
 * state already and freeing the same memory twice isn't a good
 * thing to do.
 *
 * Revision 1.55.1.2  1994/05/04  19:21:28  davidm
 * (delete_tcpstate) removed premature xFree().
 *
 * Revision 1.55.1.1  1994/04/26  17:25:34  menze
 * Added cast to bzero argument
 *
 * Revision 1.55  1994/04/21  21:09:58  davidm
 * (tcp_establishopen): use ERR_BIND instead of -1 and bzero() ex_id before
 * filling in individual fields.
 *
 * Revision 1.54  1994/01/08  21:23:50  menze
 *   [ 1994/01/05          menze ]
 *   PROTNUM changed to PORTNUM
 *
 * Revision 1.53  1993/12/16  01:45:15  menze
 * Strict ANSI compilers weren't combining multiple tentative definitions
 * of external variables into a single definition at link time.
 *
 * Revision 1.52  1993/12/13  23:26:33  menze
 * Modifications from UMass:
 *
 *   [ 93/08/20          nahum ]
 *   Fixed #undef problem.
 *
 * Revision 1.51  1993/12/11  00:25:08  menze
 * fixed #endif comments
 *
 * Revision 1.50  1993/12/07  18:15:56  menze
 * Improved validity checks on incoming participants
 *
 */

#include "xkernel.h"
#include "ip.h"
#include "tcp_internal.h"
#include "tcp_fsm.h"
#include "tcp_seq.h"
#include "tcp_timer.h"
#include "tcp_var.h"
#include "tcpip.h"
#ifdef TCP_S_IPSEC
#include "saidmgr.h"
#endif


char *prurequests[] = {
	"ATTACH",	"DETACH",	"BIND",		"LISTEN",
	"CONNECT",	"ACCEPT",	"DISCONNECT",	"SHUTDOWN",
	"RCVD",		"SEND",		"ABORT",	"CONTROL",
	"SENSE",	"RCVOOB",	"SENDOOB",	"SOCKADDR",
	"PEERADDR",	"CONNECT2",	"FASTTIMO",	"SLOWTIMO",
	"PROTORCV",	"PROTOSEND",
};


tcp_s_seq	tcp_s_iss;		/* tcp_s initial send seq # */


#ifdef __STDC__

static int	extractPart( Part *, long *, IPhost **, char * );
static void	tcp_sSessnInit( XObj );
static XObj	tcp_s_establishopen( XObj, XObj, XObj, IPhost *, IPhost *,
				   int, int, Part * );
static void	tcp_s_dumpstats( struct tcp_sstat * );
static xkern_return_t	tcp_sClose( XObj );
static int		tcp_sControlProtl( XObj, int, char *, int );
static int		tcp_sControlSessn( XObj, int, char *, int );
static XObj		tcp_sOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	tcp_sOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	tcp_sOpenDisable( XObj, XObj, XObj, Part * );
static xmsg_handle_t	tcp_sPush( XObj, Msg * );


#else

static int	extractPart();
static void	tcp_sSessnInit();
static XObj	tcp_s_establishopen();
static void	tcp_s_dumpstats();
static int		tcp_sControlProtl();
static XObj		tcp_sOpen();
static xkern_return_t	tcp_sOpenEnable();
static xkern_return_t	tcp_sOpenDisable();

#endif /* __STDC__ */



/* 
 * Check for a valid participant list
 */
#define partCheck(p, name, max, retval)					\
	{								\
	  if ( ! (p) || partLen(p) < 1 || partLen(p) > (max)) { 	\
		xTrace1(tcp_sp, TR_ERRORS,				\
			"%s -- bad participants",			\
			(name));  					\
		return (retval);					\
	  }								\
	}


/*************************************************************************/
#ifdef TCP_S_IPSEC

static xkern_return_t romuseipsec(XObj, char **, int, int, void *);

static xkern_return_t
romuseipsec(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  PSTATE     *ps = (PSTATE *)(self->state);
  ps->useIpsec=TRUE;
  return XK_SUCCESS;
}

static XObjRomOpt rmopts[1] = {{ "useipsec", 2, romuseipsec}
				};

#endif
/*************************************************************************/

void
tcp_s_init(self)
    XObj            self;
{
  Part		p;
  PSTATE	*pstate;
  struct inpcb	*tcb;

  xTrace0(tcp_sp, TR_GROSS_EVENTS, "TCP_S init");
  xTrace1(tcp_sp, TR_GROSS_EVENTS, "ENDIAN = %d", ENDIAN);

  xAssert(xIsProtocol(self));

  pstate = (PSTATE *)xMalloc(sizeof(PSTATE));
  bzero((char *)pstate,sizeof(PSTATE));
  self->state = (char *)pstate;

  self->control = tcp_sControlProtl;
  self->open = tcp_sOpen;
  self->openenable = tcp_sOpenEnable;
  self->opendisable = tcp_sOpenDisable;
  self->demux = tcp_s_input;
#ifdef TCP_S_IPSEC
  pstate->useIpsec = FALSE;
  findXObjRomOpts(self, rmopts, sizeof(rmopts)/sizeof(rmopts[0]), 0);
#endif

  pstate->passiveMap = mapCreate(37, sizeof(PassiveId));
  pstate->activeMap = mapCreate(101, sizeof(ActiveId));
  pstate->tcp_sstat = X_NEW(struct tcp_sstat);
  tcb = X_NEW(struct inpcb);
  pstate->tcb = tcb;
  tcb->inp_next = tcb->inp_prev = tcb;
  tcp_s_iss = 1;
  /* Turn on IP pseudoheader length-fixup kludge in lower protocols. */
  xControl(xGetDown(self, 0),IP_PSEUDOHDR,(char *)0,0);
  partInit(&p, 1);
  partPush(p, ANY_HOST, 0);
  xOpenEnable(self, self, xGetDown(self, 0), &p);
/*  xControl(xGetDown(self, 0), GETMYHOST, (char *)&myHost, sizeof(myHost)); */
  evDetach(evSchedule(tcp_s_fasttimo, self, TCP_S_FAST_INTERVAL));
  evDetach(evSchedule(tcp_s_slowtimo, self, TCP_S_SLOW_INTERVAL));
  tcp_sPortMapInit( &pstate->portstate );
}

/*************************************************************************/
struct tcp_sstate *new_tcp_sstate()
{
  struct tcp_sstate *tcp_sstate;
  tcp_sstate = (struct tcp_sstate *) xMalloc(sizeof(struct tcp_sstate));
  bzero((char *)tcp_sstate, sizeof(struct tcp_sstate));
  tcp_sSemInit(&tcp_sstate->waiting, 0);
  tcp_sSemInit(&tcp_sstate->lock, 1);
  tcp_sstate->closed = 0;
  tcp_sstate->snd = (struct sb *)xMalloc(sizeof(struct sb));
  /* initialize buffer size to TCP_SRCVWIN to get things started: */
  tcp_sstate->rcv_space = TCP_SRCVWIN;
  tcp_sstate->rcv_hiwat = TCP_SRCVWIN;
  return tcp_sstate;
}

/*************************************************************************/
void delete_tcp_sstate(tcp_sstate)
struct tcp_sstate *tcp_sstate;
{
  tcp_sVAll(&tcp_sstate->waiting);

  /* From RICE 07/03/90 */
  tcp_sVAll(&tcp_sstate->lock);
  xFree((char *)tcp_sstate->snd);
  xFree((char *)tcp_sstate);
}

/*************************************************************************/
static XObj
#ifdef TCP_S_IPSEC
tcp_s_establishopen(self, hlpRcv, hlpType, raddr, laddr, rport, lport, p_old)
#else
tcp_s_establishopen(self, hlpRcv, hlpType, raddr, laddr, rport, lport)
#endif
    XObj self, hlpRcv, hlpType;
    IPhost *raddr, *laddr;
    int rport, lport;
#ifdef TCP_S_IPSEC
    Part * p_old;
#endif
{
#ifdef TCP_S_IPSEC
  Part*         p;
#else
  Part          p[2];
#endif
  XObj		new;
  struct tcp_sstate *tcp_sst;
  struct inpcb *inp;
  ActiveId      ex_id;
  XObj		lls;
  PSTATE	*pstate;
  int 		x;
  int		top_sz;

  xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s_establish: on %X", hlpRcv);
  xAssert(xIsProtocol(hlpRcv));

  pstate = (PSTATE *)self->state;
  bzero((char *)&ex_id, sizeof(ex_id));
  ex_id.localport = lport;
  ex_id.remoteport = rport;
  ex_id.remoteaddr = *raddr;
#ifdef TCP_S_IPSEC
  p = (Part *) malloc(sizeof(Part) *  ((2>partLen(p_old))?2:partLen(p_old)));
#endif
  if ( mapResolve(pstate->activeMap, &ex_id, &new) == XK_SUCCESS ) {
    xTrace3(tcp_sp, TR_GROSS_EVENTS, "tcp_s_establish: (%d->%X.%d) already open",
	    ex_id.localport, ex_id.remoteaddr, ex_id.remoteport);
    return 0;
  }
#ifdef TCP_S_IPSEC
  partInit(p, ((laddr ? 2 : 1) > partLen(p_old))?
	(laddr ? 2 : 1):partLen(p_old));
#else
  partInit(p, laddr ? 2 : 1);
#endif
  partPush(p[0], raddr, sizeof(IPhost));
  if ( laddr ) {
      partPush(p[1], laddr, sizeof(IPhost));
  }
#ifdef TCP_S_IPSEC
  for(x = partLen(p_old)-1; x >1; x--)
    p[x] = p_old[x];
/*
      while ((top_sz = partStackTopByteLen(p_old[x])) >= 0)
	partPush(p[x],partPop(p_old[x]),top_sz);
*/
#endif
  xTrace0(tcp_sp, TR_EVENTS, "tcp_sOpen: opening IP");
  if ( (lls = xOpen(self, self, xGetDown(self,0), p)) == ERR_SESSN ) {
    xTrace0(tcp_sp, TR_ERRORS, "tcp_sOpen: cannot open IP session");
    return 0;
  }

  new = xCreateSessn(tcp_sSessnInit, hlpRcv, hlpType, self, 1, &lls);
  xTrace3(tcp_sp, TR_GROSS_EVENTS, "Mapping %d->%X.%d", lport, raddr, rport);
  new->binding = mapBind(pstate->activeMap,  &ex_id, new);
  if (new->binding == ERR_BIND) {
    xTrace3(tcp_sp, TR_GROSS_EVENTS, "tcp_s_establish: Bind of %d->%X.%d failed", 
      lport, raddr, rport);
    return 0;
  }

  tcp_sst = new_tcp_sstate();
  tcp_sst->hlpType = hlpType;
  new->state = (char *)tcp_sst;
  xTrace0(tcp_sp, TR_EVENTS, "tcp_sOpen: attaching...");
  tcp_s_attach(new);

  inp = sotoinpcb(new);
  tcp_sst->inpcb = inp;
  tcp_sst->tcp_scb = (struct tcp_scb *)inp->inp_ppcb;
  if (xControl(lls, GETMYHOST, (char *)&inp->inp_laddr, sizeof(inp->inp_laddr))
      <= 0) {
    xTrace3(tcp_sp, TR_ERRORS, "tcp_s_establish: getmyhost failed for 0d%d 0x%x 0d%d", 
      lport, raddr, rport);
    return 0;
  }
  inp->inp_lport = ex_id.localport;
  *(IPhost *)&inp->inp_raddr = ex_id.remoteaddr;
  inp->inp_rport = ex_id.remoteport;

  /*  xControl(xGetDown(new,0),IP_PSEUDOHDR,(char *)0,0);  moved to tcp_s_init --rcs */
  return new;
}


static int
extractPart( p, port, host, s )
    Part	*p;
    long	*port;
    IPhost	**host;
    char	*s;
{
    long	*portPtr;
    IPhost	*hostPtr;

    xAssert(port);
    if ( ! p || partStackTopByteLen(*p) < sizeof(long) ||
	 (portPtr = (long *)partPop(*p)) == 0 ) {
	xTrace1(tcp_sp, TR_SOFT_ERROR, "Bad participant in %s -- no port", s);
	return -1;
    }
    *port = *portPtr;
    if ( *port > MAX_PORT ) {
	xTrace2(tcp_sp, TR_SOFT_ERROR,
		"Bad participant in %s -- port %d out of range", s, *port);
	return -1;
    }
    if ( host ) {
	if ( partStackTopByteLen(*p) < sizeof(IPhost) ||
	     (hostPtr = (IPhost *)partPop(*p)) == 0 ) {
	    xTrace1(tcp_sp, TR_SOFT_ERROR,
		    "Bad participant in %s -- no host", s);
	    return -1;
	}
	*host = hostPtr;
    }
    return 0;
}


/*************************************************************************/
static XObj
tcp_sOpen(self, hlpRcv, hlpType, p)
    XObj            self, hlpRcv, hlpType;
    Part           *p;
{
    XObj	s;
    long	remotePort, localPort = ANY_PORT;
    IPhost	*remoteHost, *localHost = 0;
    long	freePort;
    
    xTrace0(tcp_sp, TR_GROSS_EVENTS, "TCP_S open");
#ifndef TCP_S_IPSEC
    partCheck(p, "tcp_sOpen", 2, ERR_XOBJ);
#endif
    if ( extractPart(p, &remotePort, &remoteHost, "tcp_sOpen (remote)") ) {
	return ERR_XOBJ;
    }
#ifdef TCP_S_IPSEC
    if (( partLen(p) > 1) && (partStackTopByteLen(p[1])>=0) ) {
#else
    if ( partLen(p) > 1 ) {
#endif
	if ( extractPart(p+1, &localPort, &localHost, "tcp_sOpen (local)") ) {
	    return ERR_XOBJ;
	}
	if ( localHost == (IPhost *)ANY_HOST ) {
	    localHost = 0;
	}
    }
    switch (localPort) {
    case ANY_PORT:
	/* 
	 * We need to find a free local port
	 */
	if ( tcp_sGetFreePort(((PSTATE *)(self->state))->portstate, &freePort) ) {
	    xError("tcp_sOpen -- no free ports!");
	    return ERR_XOBJ;
	}
	localPort = freePort;
	break;
    case ANY_LOW_PORT: 
	if ( tcp_sGetFreePortRange(((PSTATE *)(self->state))->portstate, 
		LOW_PORT_FLOOR, LOW_PORT_CEILING, &freePort)) {
	    xError("tcp_sOpen -- no free ports!");
	    return ERR_XOBJ;
	}
	localPort = freePort;
	break;
    case ANY_HIGH_PORT: 
	if ( tcp_sGetFreePortRange(((PSTATE *)(self->state))->portstate, 
		HIGH_PORT_FLOOR, HIGH_PORT_CEILING, &freePort)) {
	    xError("tcp_sOpen -- no free ports!");
	    return ERR_XOBJ;
	}
	localPort = freePort;
	break;
    default:
	/* 
	 * A specific local port was requested
	 */
	if ( tcp_sDuplicatePort(((PSTATE *)(self->state))->portstate, localPort))
	  {
	    xTrace1(tcp_sp, TR_MAJOR_EVENTS,
		    "tcp_sOpen: requested port %d is already in use",
		    localPort);
	    return ERR_XOBJ;
	  }
    }
    s = tcp_s_establishopen(self, hlpRcv, hlpType, remoteHost, localHost,
#ifdef TCP_S_IPSEC
			  remotePort, localPort, p);
#else
			  remotePort, localPort);
#endif
    if (!s) {
	tcp_sReleasePort(((PSTATE *)(self->state))->portstate, localPort);
	s = ERR_SESSN;
    } else {
	xTrace0(tcp_sp, TR_EVENTS, "tcp_sOpen: connecting...");
	tcp_s_usrreq(s, PRU_CONNECT, 0, 0);
	
	xTrace0(tcp_sp, TR_EVENTS, "tcp_sOpen: waiting...");
	tcp_sSemWait(&(sototcp_sst(s)->waiting));
	if (sototcp_scb(s)->t_state != TCP_SS_ESTABLISHED) {
	    xTrace3(tcp_sp, TR_EVENTS, "tcp_sOpen: open (%d->%X.%d) connect failed",
		    localPort,  ipHostStr(remoteHost), remotePort);
	    tcp_sReleasePort(((PSTATE *)(self->state))->portstate, localPort);
	    s = ERR_SESSN;
	}
    }
    xTrace1(tcp_sp, TR_GROSS_EVENTS, "Return from tcp_sOpen, session = %x", s);
    return s;
}

/*************************************************************************/
static xkern_return_t
tcp_sOpenEnable(self, hlpRcv, hlpType, p)
    XObj	self, hlpRcv, hlpType;
    Part	*p;
{
    PassiveId	key;
    long	protNum;
    PSTATE	*pstate;
    Enable	*e;
    
    partCheck(p, "tcp_sOpenEnable", 1, XK_FAILURE);
    if ( extractPart(p, &protNum, 0, "tcp_sOpenEnable") ) {
	return XK_FAILURE;
    }
    key = protNum;
    pstate = (PSTATE *)self->state;
    xTrace2(tcp_sp, TR_MAJOR_EVENTS, "tcp_sOpenEnable mapping %d->%x", key,
	    hlpRcv);
    if ( mapResolve(pstate->passiveMap, &key, &e) == XK_SUCCESS ) {
	if ( e->hlpRcv == hlpRcv && e->hlpType == hlpType ) {
	    e->rcnt++;
	    return XK_SUCCESS;
	}
	return XK_FAILURE;
    }
    tcp_sDuplicatePort(pstate->portstate, key);
    e = (Enable *)xMalloc(sizeof(Enable));
    e->hlpRcv = hlpRcv;
    e->hlpType = hlpType;
    e->rcnt = 1;
    e->binding = mapBind(pstate->passiveMap, &key, e);
    if ( e->binding == ERR_BIND ) {
	xFree((char *)e);
	return XK_FAILURE;
    }
    return XK_SUCCESS;
}

/*************************************************************************/
static xkern_return_t
tcp_sOpenDisable(self, hlpRcv, hlpType, p)
    XObj	self, hlpRcv, hlpType;
    Part        *p;
{
    PassiveId	key;
    long	protNum;
    Enable	*e;
    PSTATE	*pstate;

    pstate = (PSTATE *)self->state;
    partCheck(p, "tcp_sOpenDisable", 1, XK_FAILURE);
    if ( extractPart(p, &protNum, 0, "tcp_sOpenEnable") ) {
	return XK_FAILURE;
    }
    key = protNum;
    xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s_disable removing %d", key);
    if ( mapResolve(pstate->passiveMap, &key, &e) == XK_FAILURE ||
		e->hlpRcv != hlpRcv || e->hlpType != hlpType ) {
	return XK_FAILURE;
    }
    if (--(e->rcnt) == 0) {
	mapRemoveBinding(pstate->passiveMap, e->binding);
	xFree((char *)e);
	tcp_sReleasePort(pstate->portstate, key);
    }
    return XK_SUCCESS;
}

/*************************************************************************/
/*
 * User issued close, and wish to trail through shutdown states:
 * if never received SYN, just forget it.  If got a SYN from peer,
 * but haven't sent FIN, then go to FIN_WAIT_1 state to send peer a FIN.
 * If already got a FIN from peer, then almost done; go to LAST_ACK
 * state.  In all other cases, have already sent FIN to peer (e.g.
 * after PRU_SHUTDOWN), and just have to play tedious game waiting
 * for peer to send FIN or not respond to keep-alives, etc.
 * We can let the user exit from the close as soon as the FIN is acked.
 */
static xkern_return_t
tcp_sClose(so)
    XObj	so;
{
    register struct tcp_scb *tp;
    register struct tcp_sstate *tcp_sst;
    
    xTrace1(tcp_sp, TR_FUNCTIONAL_TRACE, "tcp_sClose(so=%08x)", so);

    tcp_sst = sototcp_sst(so);
    tcp_sst->closed |= 1;
    
    tp = sototcp_scb(so);
    tp = tcp_s_usrclosed(tp);
    if (tp) tcp_s_output(tp);

    xTrace2(tcp_sp, TR_MAJOR_EVENTS, "tcp_sClose: tp=%08x, tcp_sst->closed=%x",
	    tp, tcp_sst->closed);
    
    if (tcp_sst->closed == 1) {
	/*
	 * This was a user close before the network close
	 */
	xTrace0(tcp_sp, TR_EVENTS, "tcp_sClose: waiting...");
	tcp_sSemWait(&tcp_sst->waiting);
	xTrace0(tcp_sp, TR_EVENTS, "tcp_sClose: done");
    }
    return XK_SUCCESS;
}


/*************************************************************************/
static xmsg_handle_t
tcp_sPush(so, msg)
     XObj so;
     Msg *msg;
{
    int error;
    int space;
    Msg pushMsg;
    register struct tcp_sstate *tcp_sst = sototcp_sst(so);
    register struct tcp_scb *tp = sototcp_scb(so);

    xTrace2(tcp_sp, TR_MAJOR_EVENTS, "tcp_sPush: session %X, msg %d bytes",
	    so, msgLen(msg));
    if (so->rcnt < 1) {
	xTrace1(tcp_sp, TR_GROSS_EVENTS, "tcp_sPush: session ref count = %d",
		so->rcnt);
	return XMSG_ERR_HANDLE;
    }
    if (tcp_sst->closed) {
	xTrace1(tcp_sp, TR_SOFT_ERROR, "tcp_sPush: session already closed %d",
		tcp_sst->closed);
	return XMSG_ERR_HANDLE;
    }

#undef CHUNKSIZE /* (sbhiwat(tcp_sst->snd) / 2) */
#define CHUNKSIZE (tp->t_maxseg)

    if ((tcp_sst->flags & TCP_SST_NBIO) &&
	(sbspace(tcp_sst->snd) < MIN(msgLen(msg), CHUNKSIZE)))
    {
	return XMSG_ERR_WOULDBLOCK;
    } /* if */

    msgConstructEmpty(&pushMsg);
    while ( msgLen(msg) != 0 ) {
	xTrace1(tcp_sp, TR_FUNCTIONAL_TRACE,
		"tcp_sPush: msgLen = <%d>", msgLen(msg));
	msgChopOff(msg, &pushMsg, CHUNKSIZE);
	xTrace2(tcp_sp, TR_FUNCTIONAL_TRACE,
		"tcp_sPush: after break pushMsg = <%d> msg = <%d>",
		msgLen(&pushMsg), msgLen(msg));
	while ((space = sbspace(tcp_sst->snd)) < msgLen(&pushMsg)) {
	    xTrace1(tcp_sp, TR_GROSS_EVENTS,
		    "tcp_sPush: waiting for space (%d available)", space);
	    xTrace1(tcp_sp, TR_FUNCTIONAL_TRACE,
		    "tcp_sPush: msgLen == %d before blocking",
		    msgLen(msg));
	    tcp_sSemWait(&tcp_sst->waiting);
	    if (tcp_sst->closed) {
		xTrace1(tcp_sp, TR_FUNCTIONAL_TRACE,
			"tcp_sPush: session already closed %d", tcp_sst->closed);
		msgDestroy(&pushMsg);
		return XMSG_ERR_HANDLE;
	    }
	}
	xTrace1(tcp_sp, TR_FUNCTIONAL_TRACE,
		"tcp_spush about to append to sb.  orig msgLen == %d",
		msgLen(msg));
	sbappend(tcp_sst->snd, &pushMsg);
	if ( tcp_sst->flags & TCP_SST_PUSH_ALWAYS ) {
	    tp->t_force = 1;
	    error = tcp_s_output(sototcp_scb(so));
	    tp->t_force = 0;
	} else {
	    error = tcp_s_output(sototcp_scb(so));
	}
	if (error) {
	    xTrace1(tcp_sp, TR_ERRORS, "tcp_sPush failed with code %d", error);
	    msgDestroy(&pushMsg);
	    return XMSG_ERR_HANDLE;
	}
    }
    msgDestroy(&pushMsg);
    return XMSG_NULL_HANDLE;
}

/*************************************************************************/
static int
tcp_sControlSessn(s, opcode, buf, len)
    Sessn           s;
    char           *buf;
    int             opcode;
    int             len;
{
  struct tcp_sstate *ts;
  struct tcp_scb   *tp;
  u_short size;

  ts = (struct tcp_sstate *) s->state;
  tp = ts->tcp_scb;
  switch (opcode) {

   case GETMYPROTO:
     checkLen(len, sizeof(long));
     *(long *)buf = tp->t_template->ti_sport;
     return sizeof(long);

   case GETPEERPROTO:
     checkLen(len, sizeof(long));
     *(long *)buf = tp->t_template->ti_dport;
     return sizeof(long);

   case TCP_PUSH:
     tp->t_force = 1;
     tcp_s_output(tp);
     tp->t_force = 0;
     return 0;

   case TCP_GETSTATEINFO:
     *(int *) buf = tp->t_state;
     return (sizeof(int));

   case GETOPTPACKET:
   case GETMAXPACKET:
     if ( xControl(xGetDown(s, 0), opcode, buf, len) < sizeof(int) ) {
	 return -1;
     }
     *(int *)buf -= sizeof(struct tcp_shdr);
     return sizeof(int);
     
   case GETPARTICIPANTS:
     {
	 Part		p[2];
	 int 		retLen;
	 long		localPort, remotePort;
	 int		numParts;
	 xkern_return_t	xkr;
	 
	 retLen = xControl(xGetDown(s, 0), opcode, buf, len);
	 if ( retLen < 0 ) {
	     return retLen;
	 }
	 numParts = partExtLen(buf);
	 if ( numParts > 0 && numParts <= 2 ) {
	     partInternalize(p, buf);
	     if ( numParts == 2 ) {
		 localPort = tp->t_template->ti_sport;
		 partPush(p[1], &localPort, sizeof(long));
	     }
	     remotePort = tp->t_template->ti_dport;
	     partPush(p[0], &remotePort, sizeof(long));
	     xkr = partExternalize(p, buf, &len);
	     if ( xkr == XK_FAILURE ) {
		 xTrace0(tcp_sp, TR_SOFT_ERRORS,
			 "TCP_S GETPARTS -- buffer too small");
		 retLen = -1;
	     } else {
		 retLen = len;
	     }
	 } else {
	     xTrace1(tcp_sp, TR_SOFT_ERRORS,
		 "TCP_S -- Bad number of participants (%d) returned in GETPARTS",
		     partExtLen(buf));
	     retLen = -1;
	 }
	 return retLen;
     }

   case SETNONBLOCKINGIO:
     xControl(xGetDown(s, 0), opcode, buf, len);
     if (*(int*)buf) {
	 ts->flags |= TCP_SST_NBIO;
     } else {
	 ts->flags &= ~TCP_SST_NBIO;
     } /* if */
     return 0;

   case TCP_SETPUSHALWAYS:
     checkLen(len, sizeof(int));
     if (*(int*)buf) {
	 ts->flags |= TCP_SST_PUSH_ALWAYS;
     } else {
	 ts->flags &= ~TCP_SST_PUSH_ALWAYS;
     } /* if */
     return 0;

   case TCP_SETRCVACKALWAYS:
     checkLen(len, sizeof(int));
     if (*(int*)buf) {
	 ts->flags |= TCP_SST_RCV_ACK_ALWAYS;
     } else {
	 ts->flags &= ~TCP_SST_RCV_ACK_ALWAYS;
     } /* if */
     return 0;

   case TCP_SETRCVBUFSPACE:
     checkLen(len, sizeof(u_short));
     size = *(u_short *)buf;
     ts->rcv_space += size;   /* changed for TCP_S Forwarder */
     /* after receiving message possibly send window update: */
     tcp_s_output(tp);
     /*     printf("Receive window space: %d\n", ts->rcv_space);*/
     return 0;

   case TCP_GETSNDBUFSPACE:
     checkLen(len, sizeof(u_short));
     *(u_short*)buf = sbspace(ts->snd);
     return sizeof(u_short);

   case TCP_SETRCVBUFSIZE:
     checkLen(len, sizeof(u_short));
     size = *(u_short *)buf;
     ts->rcv_hiwat = size;
     return 0;

   case TCP_SETSNDBUFSIZE:
     checkLen(len, sizeof(u_short));
     sbhiwat(ts->snd) = *(u_short *)buf;
     return 0;

   case TCP_SETOOBINLINE:
     checkLen(len, sizeof(int));
     if (*(int*)buf) {
	 ts->flags |= TCP_SST_OOBINLINE;
     } else {
	 ts->flags &= ~TCP_SST_OOBINLINE;
     } /* if */
     return 0;

   case TCP_GETOOBDATA:
     {     
	 char peek;

	 checkLen(len, sizeof(char));

	 peek = *(char*)buf;
	 if ((tp->t_oobflags & TCP_SOOB_HADDATA) ||
	     !(tp->t_oobflags & TCP_SOOB_HAVEDATA))
	 {
	     return 0;
	 } /* if */
	 *(char*)buf = tp->t_iobc;
	 if (!peek) {
	     tp->t_oobflags ^= TCP_SOOB_HAVEDATA | TCP_SOOB_HADDATA;
	 } /* if */
	 return 1;
     }

   case TCP_OOBPUSH:
     {
	 Msg *m;

	 checkLen(len, sizeof(Msg*));

	 m = (Msg*)buf;
	 sbappend(ts->snd, m);
	 tp->snd_up = tp->snd_una + sblength(ts->snd);
	 tp->t_force = 1;
	 tcp_s_output(tp);
	 tp->t_force = 0;
	 return 0;
     }

   default:
     return xControl(xGetDown(s, 0), opcode, buf, len);
  }
}

/*************************************************************************/
static int
tcp_sControlProtl(self, opcode, buf, len)
    XObj	    self;
    int             opcode, len;
    char           *buf;
{
    long port;

    switch (opcode) {

      case TCP_DUMPSTATEINFO:
	{
	    PSTATE	*ps = (PSTATE *)self->state;

	    tcp_s_dumpstats(ps->tcp_sstat);
	    return 0;
	}

      case TCP_GETFREEPORTNUM:
	checkLen(len, sizeof(long));
	port = *(long *)buf;
	if (tcp_sGetFreePort(((PSTATE *)(self->state))->portstate, &port)) {
	    return -1;
	} /* if */
	*(long *)buf = port;
	return 0;

      case TCP_RELEASEPORTNUM:
	checkLen(len, sizeof(long));
	port = *(long *)buf;
	tcp_sReleasePort(((PSTATE *)(self->state))->portstate, port);
	return 0;

      default:
	return xControl(xGetDown(self,0), opcode, buf, len);
    }
}

/*************************************************************************/
static void
tcp_sSessnInit(s)
    XObj s;
{
    xAssert(xIsSession(s));
    s->push = tcp_sPush;
    s->close = tcp_sClose;
    s->control = tcp_sControlSessn;
    s->demux = tcp_s_input;
    s->pop = tcp_sPop;
}

/*************************************************************************/
void
sorwakeup(so)
XObj so;
{
  xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s: sorwakeup on %X", so);
}
    
/*************************************************************************/
void
sowwakeup(so)
XObj so;
{
  xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s: sowwakeup on %X", so);
  tcp_sSemSignal(&sototcp_sst(so)->waiting);
}

/*************************************************************************/
/*ARGSUSED*/
int
soreserve(so, send, recv)
XObj so;
int send, recv;
{
  struct tcp_sstate *tcp_sst = sototcp_sst(so);
  sbinit(tcp_sst->snd);
  return 0;
}

/*************************************************************************/
void
socantsendmore(so)
    XObj so;
{
    xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s:  socantsendmore on %X", so);
}

/*************************************************************************/
void
soisdisconnected(so)
    XObj so;
{
    xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s:  soisdisconnected on %X", so);
    if (sototcp_sst(so)->closed) {
	/* It was already closed, either by the network or the user */
    } else {
	/* deliver close done after we're done handling current message: */
	xCloseDone(so);
    } /* if */
    tcp_sVAll(&sototcp_sst(so)->waiting);
} /* soisdisconnected */

/*************************************************************************/
void
soabort(so)
XObj so;
{
}

/*************************************************************************/
void
socantrcvmore(so)
    XObj so;
{
    xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s:  socantrcvmore on %X", so);
    
    sototcp_sst(so)->closed |= 2;
    if (sototcp_sst(so)->closed & 1) {
	/* do nothing, the user already knows that it is closed */
    } else {
	xCloseDone(so);
    }
}

/*************************************************************************/
void
soisconnected(so)
    XObj so;
{
    struct tcp_sstate	*state = sototcp_sst(so);
    Tcp_SSem 		*s = &state->waiting;

    xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s:  soisconnected on %X", so);
    if (s->waitCount) {
	xTrace1(tcp_sp, TR_EVENTS, "tcp_s:  waking up opener on %X", so);
	tcp_sSemSignal(s);
    } else {
	/*
	 * I think that this is the time for a open done
	 */
	xOpenDone(xGetUp(so), so, xMyProtl(so));
    }
}

/*************************************************************************/
void
soisconnecting(so)
    XObj so;
{
    xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s:  soisconnecting on %X", so);
}

/*************************************************************************/
void
soisdisconnecting(so)
XObj so;
{
    xTrace1(tcp_sp, TR_MAJOR_EVENTS, "tcp_s:  soisdisconnecting on %X", so);
}

/*************************************************************************/
void
xGetParticipants(s, p, n, b)
XObj s;
Part *p;
char **b;
{
  char *buf = xMalloc(sizeof(Part)*n);
  xControl(s, GETPARTICIPANTS, buf, sizeof(Part)*n);
  partInternalize(p, buf);
  xFree(buf);
}
/*************************************************************************/
/* 
 * sonewconn is called by the input routine to establish a passive open
 */
XObj
#ifdef TCP_S_IPSEC
sonewconn(self, so, hlpType, transport_s, src, dst, sport, dport)
    XObj self, so, hlpType, transport_s;
#else
sonewconn(self, so, hlpType, src, dst, sport, dport)
    XObj self, so, hlpType;
#endif
    IPhost *src, *dst;
    int sport, dport;
{
  XObj new;
#ifdef TCP_S_IPSEC
  Part p[SPI_PART+1];
  char *pbuf = 0;;
#ifndef IPSEC
  PSTATE	*pstate = (PSTATE *) self->state;
  XSaid lsaid;
  if ((pstate->useIpsec) &&
	 ((lsaid = xControl(transport_s,SAIDMGR_GETLSAID,NULL,0)) >= 0))
  {
    	partInit(p,SPI_PART+1);
  	partPush(p[SPI_PART],&lsaid,sizeof(lsaid));
  }
  else
  {
	partInit(p,1);
  }
#else /* IPSEC */
  xGetParticipants(transport_s, p, SPI_PART+1, &pbuf);
#endif /* IPSEC */
#endif
  xAssert(xIsProtocol(so));
  xTrace1(tcp_sp, TR_MAJOR_EVENTS, "sonewconn on %X", so);
#ifdef TCP_S_IPSEC
  new = tcp_s_establishopen(self, so, hlpType, src, dst, sport, dport, p);
#ifdef IPSEC
  if (pbuf) xFree(pbuf);
#endif /* IPSEC */
#else
  new = tcp_s_establishopen(self, so, hlpType, src, dst, sport, dport);
#endif
  if ( new ) {
      tcp_sDuplicatePort(((PSTATE *)self->state)->portstate, dport);
  }
  return new;
}


/*************************************************************************/
/* 
 * sohasoutofband is called by the input routine to signal the presence
 * of urgent (out-of-band) data
 */
void
sohasoutofband(so, oobmark)
     XObj so;
     u_int oobmark;
{
    void *buf[2];

    buf[0] = so;
    buf[1] = (void*) oobmark;
    xControl(xGetUp(so), TCP_OOBMODE, (char*) buf, sizeof(buf));
} /* sohasoutofband */

/*************************************************************************/

void
tcp_sSemWait( ts )
    Tcp_SSem	*ts;
{
    ts->waitCount++;
    semWait(&ts->s);
    ts->waitCount--;
}


void
tcp_sSemSignal( ts )
    Tcp_SSem	*ts;
{
    semSignal(&ts->s);
}


void
tcp_sVAll( ts )
    Tcp_SSem	*ts;
{
    int i, n;

    n=ts->waitCount;
    for ( i=0; i < n; i++ ) {
	semSignal(&ts->s);
    }
}


void
tcp_sSemInit( ts, n )
    Tcp_SSem	*ts;
    int		n;
{
    semInit(&ts->s, n);
    ts->waitCount = 0;
}

/*************************************************************************/
static void
tcp_s_dumpstats( tcp_sstat )
    struct tcp_sstat	*tcp_sstat;
{
  printf("tcp_ss_badsum %d\n", tcp_sstat->tcp_ss_badsum);
  printf("tcp_ss_badoff %d\n", tcp_sstat->tcp_ss_badoff);
  printf("tcp_ss_hdrops %d\n", tcp_sstat->tcp_ss_hdrops);
  printf("tcp_ss_badsegs %d\n", tcp_sstat->tcp_ss_badsegs);
  printf("tcp_ss_unack %d\n", tcp_sstat->tcp_ss_unack);
  printf("connections initiated %d\n", tcp_sstat->tcp_ss_connattempt);
  printf("connections accepted %d\n", tcp_sstat->tcp_ss_accepts);
  printf("connections established %d\n", tcp_sstat->tcp_ss_connects);
  printf("connections dropped %d\n", tcp_sstat->tcp_ss_drops);
  printf("embryonic connections dropped %d\n", tcp_sstat->tcp_ss_conndrops);
  printf("conn. closed (includes drops) %d\n", tcp_sstat->tcp_ss_closed);
  printf("segs where we tried to get rtt %d\n", tcp_sstat->tcp_ss_segstimed);
  printf("times we succeeded %d\n", tcp_sstat->tcp_ss_rttupdated);
  printf("delayed acks sent %d\n", tcp_sstat->tcp_ss_delack);
  printf("conn. dropped in rxmt timeout %d\n", tcp_sstat->tcp_ss_timeoutdrop);
  printf("retransmit timeouts %d\n", tcp_sstat->tcp_ss_rexmttimeo);
  printf("persist timeouts %d\n", tcp_sstat->tcp_ss_persisttimeo);
  printf("keepalive timeouts %d\n", tcp_sstat->tcp_ss_keeptimeo);
  printf("keepalive probes sent %d\n", tcp_sstat->tcp_ss_keepprobe);
  printf("connections dropped in keepalive %d\n", tcp_sstat->tcp_ss_keepdrops);
  printf("total packets sent %d\n", tcp_sstat->tcp_ss_sndtotal);
  printf("data packets sent %d\n", tcp_sstat->tcp_ss_sndpack);
  printf("data bytes sent %d\n", tcp_sstat->tcp_ss_sndbyte);
  printf("data packets retransmitted %d\n", tcp_sstat->tcp_ss_sndrexmitpack);
  printf("data bytes retransmitted %d\n", tcp_sstat->tcp_ss_sndrexmitbyte);
  printf("ack-only packets sent %d\n", tcp_sstat->tcp_ss_sndacks);
  printf("window probes sent %d\n", tcp_sstat->tcp_ss_sndprobe);
  printf("packets sent with URG only %d\n", tcp_sstat->tcp_ss_sndurg);
  printf("window update-only packets sent %d\n", tcp_sstat->tcp_ss_sndwinup);
  printf("control (SYN|FIN|RST) packets sent %d\n", tcp_sstat->tcp_ss_sndctrl);
  printf("total packets received %d\n", tcp_sstat->tcp_ss_rcvtotal);
  printf("packets received in sequence %d\n", tcp_sstat->tcp_ss_rcvpack);
  printf("bytes received in sequence %d\n", tcp_sstat->tcp_ss_rcvbyte);
  printf("packets received with ccksum errs %d\n", tcp_sstat->tcp_ss_rcvbadsum);
  printf("packets received with bad offset %d\n", tcp_sstat->tcp_ss_rcvbadoff);
  printf("packets received too short %d\n", tcp_sstat->tcp_ss_rcvshort);
  printf("duplicate-only packets received %d\n", tcp_sstat->tcp_ss_rcvduppack);
  printf("duplicate-only bytes received %d\n", tcp_sstat->tcp_ss_rcvdupbyte);
  printf("packets with some duplicate data %d\n", tcp_sstat->tcp_ss_rcvpartduppack);
  printf("dup. bytes in part-dup. packets %d\n", tcp_sstat->tcp_ss_rcvpartdupbyte);
  printf("out-of-order packets received %d\n", tcp_sstat->tcp_ss_rcvoopack);
  printf("out-of-order bytes received %d\n", tcp_sstat->tcp_ss_rcvoobyte);
  printf("packets with data after window %d\n", tcp_sstat->tcp_ss_rcvpackafterwin);
  printf("bytes rcvd after window %d\n", tcp_sstat->tcp_ss_rcvbyteafterwin);
  printf("packets rcvd after \"close\" %d\n", tcp_sstat->tcp_ss_rcvafterclose);
  printf("rcvd window probe packets %d\n", tcp_sstat->tcp_ss_rcvwinprobe);
  printf("rcvd duplicate acks %d\n", tcp_sstat->tcp_ss_rcvdupack);
  printf("rcvd acks for unsent data %d\n", tcp_sstat->tcp_ss_rcvacktoomuch);
  printf("rcvd ack packets %d\n", tcp_sstat->tcp_ss_rcvackpack);
  printf("bytes acked by rcvd acks %d\n", tcp_sstat->tcp_ss_rcvackbyte);
  printf("rcvd window update packets %d\n", tcp_sstat->tcp_ss_rcvwinupd);


  tcp_sstat->tcp_ss_badsum = 0;
  tcp_sstat->tcp_ss_badoff = 0;
  tcp_sstat->tcp_ss_hdrops = 0;
  tcp_sstat->tcp_ss_badsegs = 0;
  tcp_sstat->tcp_ss_unack = 0;
  tcp_sstat->tcp_ss_connattempt = 0;
  tcp_sstat->tcp_ss_accepts = 0;
  tcp_sstat->tcp_ss_connects = 0;
  tcp_sstat->tcp_ss_drops = 0;
  tcp_sstat->tcp_ss_conndrops = 0;
  tcp_sstat->tcp_ss_closed = 0;
  tcp_sstat->tcp_ss_segstimed = 0;
  tcp_sstat->tcp_ss_rttupdated = 0;
  tcp_sstat->tcp_ss_delack = 0;
  tcp_sstat->tcp_ss_timeoutdrop = 0;
  tcp_sstat->tcp_ss_rexmttimeo = 0;
  tcp_sstat->tcp_ss_persisttimeo = 0;
  tcp_sstat->tcp_ss_keeptimeo = 0;
  tcp_sstat->tcp_ss_keepprobe = 0;
  tcp_sstat->tcp_ss_keepdrops = 0;
  tcp_sstat->tcp_ss_sndtotal = 0;
  tcp_sstat->tcp_ss_sndpack = 0;
  tcp_sstat->tcp_ss_sndbyte = 0;
  tcp_sstat->tcp_ss_sndrexmitpack = 0;
  tcp_sstat->tcp_ss_sndrexmitbyte = 0;
  tcp_sstat->tcp_ss_sndacks = 0;
  tcp_sstat->tcp_ss_sndprobe = 0;
  tcp_sstat->tcp_ss_sndurg = 0;
  tcp_sstat->tcp_ss_sndwinup = 0;
  tcp_sstat->tcp_ss_sndctrl = 0;
  tcp_sstat->tcp_ss_rcvtotal = 0;
  tcp_sstat->tcp_ss_rcvpack = 0;
  tcp_sstat->tcp_ss_rcvbyte = 0;
  tcp_sstat->tcp_ss_rcvbadsum = 0;
  tcp_sstat->tcp_ss_rcvbadoff = 0;
  tcp_sstat->tcp_ss_rcvshort = 0;
  tcp_sstat->tcp_ss_rcvduppack = 0;
  tcp_sstat->tcp_ss_rcvdupbyte = 0;
  tcp_sstat->tcp_ss_rcvpartduppack = 0;
  tcp_sstat->tcp_ss_rcvpartdupbyte = 0;
  tcp_sstat->tcp_ss_rcvoopack = 0;
  tcp_sstat->tcp_ss_rcvoobyte = 0;
  tcp_sstat->tcp_ss_rcvpackafterwin = 0;
  tcp_sstat->tcp_ss_rcvbyteafterwin = 0;
  tcp_sstat->tcp_ss_rcvafterclose = 0;
  tcp_sstat->tcp_ss_rcvwinprobe = 0;
  tcp_sstat->tcp_ss_rcvdupack = 0;
  tcp_sstat->tcp_ss_rcvacktoomuch = 0;
  tcp_sstat->tcp_ss_rcvackpack = 0;
  tcp_sstat->tcp_ss_rcvackbyte = 0;
  tcp_sstat->tcp_ss_rcvwinupd = 0;
}
